#!/usr/bin/env python

import getopt, sys, logging, os
import ConfigParser

from parsers.ModelParser import ModelParser
from parsers.InformationParser import InformationParser
from parsers.RuleParser import RuleParser
from generators.PrologGenerator import PrologGenerator

def main():
    """Start the program for verification
    
    This is a doctest sample
    >>> usage()
    RUBACON -- Rule Based Access Control
    ---------------------------------------
    Usage rubacon.py [options]:
         -h, --help          Print this help screen
         -M, --model         Load the model in the given file
         -I, --information   Load the information in the given file
         -R, --rules         Load the rules in the given file
         -o, --output=       Write output to file
         -v                  Verbose output
    """
    # Default is to write out warnings only.
    #logging.basicConfig(level=logging.DEBUG,
    #    format='%(asctime)s %(levelname)-8s %(message)s')
    config_file = "rubacon.cfg"
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hM:I:R:Cv:o:c:P:", ["help", "model=", "info=", "rules=", "clear", "verbose=", "output=", "config_file=", "prolog="])
    except getopt.GetoptError:
        usage()
        sys.exit(2)
    output = None
    verbose = False
    
    #
    #    Treats the options first.
    #
    for o,a in opts:
        if o in ("-v", "--verbose"):
            verbose = True
            a = a.lower()
            if a not in ("debug", "info", "warning" ,"error"):
                usage()
                sys.exit(2)
            if a.lower() == "debug":
                logging.basicConfig(level=logging.DEBUG,
                    format='%(asctime)s %(levelname)-8s %(message)s')
                logging.debug("Setting verbosity to level DEBUG.")
            elif a.lower() == "info":
                logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s %(levelname)-8s %(message)s')
                logging.info("Setting verbosity to level INFO.")
            elif a.lower() == "warning":
                logging.basicConfig(level=logging.WARNING,
                    format='%(asctime)s %(levelname)-8s %(message)s')
                logging.warning("Setting verbosity to level WARNING.")
            else:
                logging.basicConfig(level=logging.ERROR,
                    format='%(asctime)s %(levelname)-8s %(message)s')
                logging.warning("Setting verbosity to level ERROR.")
        
        if o in ("-c", "--config_file"):
            config_file = a
            if not os.path.exists(a):
                logging.error("Config file at '%s' not found." % a)
                sys.exit(2)
            logging.debug("Setting configuration to: '%s'" % a)
        if o in ("-h", "--help"):
            usage()
            sys.exit()
        if o in ("-o", "--output"):
            output = a
    
    # Create a default level
    if not verbose:
        logging.basicConfig(level=logging.INFO,
            format='%(asctime)s %(levelname)-8s %(message)s')  
    
    # Read the config file at the default position:
    config = ConfigParser.ConfigParser()
    config.readfp(open('rubacon/defaults.cfg'))
    config.read([config_file, 'rubacon.cfg', os.path.expanduser('~/.rubacon.cfg')])
    
         
    for (o, a) in opts:
        #
        #    Clean the storage and write a new model to it.
        #
        if o in ("-M", "--model"):
            if not os.path.exists(a):
                logging.error("Model file not found: '%s'" % a)
                sys.exit(2)
            model_storage = config.get('model', 'storage')

            logging.info("Write model file at '%s' to the storage at %s." % (a, model_storage))
            
            # Remove the old model if it exists:
            if (os.path.exists(model_storage)):
                os.remove(model_storage)
            ModelParser.parse(a, config=config)
        
        #
        #    Write the information to the storage.
        #    There is not automatic cleaning, since this
        #    is intended to be updated by split info files.
        #
        if o in ("-I", "--information"):
            if not os.path.exists(a):
                logging.error("Information file not found: '%s'" % a)
                sys.exit(2)
            info_storage = config.get('information', 'storage')
            logging.info("Write information file at '%s' to the storage at '%s'" % (a, info_storage))
            InformationParser.parse(a, config=config)
        
        #
        #    Read the rules from the file and store
        #    them in the given position.
        #  
        if o in ("-R", "--rules"):
            if not os.path.exists(a):
                logging.error("Rule file not found: %s" % a)
                sys.exit(2)
            rule_storage = config.get('rules', 'storage')
            logging.info("Write rules file at '%s' to the storage at '%s'" % (a, rule_storage))
            RuleParser.parse(a, config=config)
            
        if o in ("-C", "--clean"):
            if os.path.exists(config.get('model', 'storage')):
                os.remove(config.get('model', 'storage'))
            if os.path.exists(config.get('information', 'storage')):
                os.remove(config.get('information', 'storage'))
            if os.path.exists(config.get('rules', 'storage')):
                os.remove(config.get('rules', 'storage'))
        
        if o in ("-P", "--prolog"):
            if os.path.exists(a):
                logging.error("Output file '%s' already exists!" % a)
                sys.exit(2)
            generator = PrologGenerator()
            generator.generate(a, config=config)
    pass

def usage():
    """Print usage info for the file
    
    """
    print """RUBACON -- Rule Based Access Control
---------------------------------------
Usage rubacon.py [options]:
     -h, --help          Print this help screen
     -M, --model         Load the model in the given file
     -I, --information   Load the information in the given file
     -R, --rules         Load the rules in the given file
     -o, --output=       Write output to file
     -v, --verbose=      Verbosity level output, one of DEBUG, INFO, WARNING"""
     
if __name__ == "__main__":
    # import doctest
    # doctest.testmod()
    main()